// Copyright (c) 2019 Uber Technologies, Inc.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

import React, {Component} from 'react';
import styled from 'styled-components';
import PropTypes from 'prop-types';
import {sortable} from 'react-anything-sortable';

import WidgetConfigurator from './widget-configurator';
import WidgetPanelHeader from './widget-panel-header';
// import layers from '../../../../dist/components/common/icons/layers';

const PanelWrapper = styled.div`
  font-size: 12px;
  border-radius: 1px;
  margin-bottom: 8px;

  &.dragging {
    cursor: move;
  }
`;
const StyledWidgetContent = styled.div`
  background-color: ${props => props.theme.panelBackground};
  padding: 12px;
  font-size: 24px;
`;

function WidgetPanelFactory() {
  @sortable
  class WidgetPanel extends Component {
    static propTypes = {
      widget: PropTypes.object.isRequired,
      datasets: PropTypes.object.isRequired,
      idx: PropTypes.number.isRequired,
      widgetConfigChange: PropTypes.func.isRequired,
      widgetTypeChange: PropTypes.func.isRequired,
      // layers: propTypes.arrayOf(propTypes.any),
      openModal: PropTypes.func.isRequired,
      removeWidget: PropTypes.func.isRequired,
      onCloseConfig: PropTypes.func,

      widgetTypeOptions: PropTypes.arrayOf(PropTypes.any),
      widgetVisConfigChange: PropTypes.func,
      widgetVisualChannelConfigChange: PropTypes.func
    };

    updateWidgetConfig = newProp => {
      this.props.widgetConfigChange(this.props.widget, newProp);
    };

    updateWidgetType = newType => {
      this.props.widgetTypeChange(this.props.widget, newType);
    };

    updateWidgetVisConfig = newVisConfig => {
      this.props.widgetVisConfigChange(this.props.widget, newVisConfig);
    };

    updateWidgetVisualChannelConfig = (newConfig, channel, scaleKey) => {
      this.props.widgetVisualChannelConfigChange(
        this.props.widget,
        newConfig,
        channel,
        scaleKey
      );
    };

    _updateWidgetLabel = ({target: {value}}) => {
      this.updateWidgetConfig({label: value});
    };

    _toggleVisibility = e => {
      e.stopPropagation();
      const isVisible = !this.props.widget.config.isVisible;
      this.updateWidgetConfig({isVisible});
    };

    _toggleEnableConfig = e => {
      e.stopPropagation();
      const {
        widget: {
          config: {isConfigActive}
        }
      } = this.props;
      this.updateWidgetConfig({isConfigActive: !isConfigActive});
    };

    _removeWidget = e => {
      e.stopPropagation();
      this.props.removeWidget(this.props.idx);
    };
    render() {
      const {widget, layers, idx, datasets, widgetTypeOptions} = this.props;
      const {config} = widget;
      // console.log('widget', widget);
      // console.log('widgetPanel datasetsColor', config.dataId);
      const {isConfigActive} = config;
      // console.log('configInWidgetPanel', config);
      return (
        <PanelWrapper
          active={isConfigActive}
          className={`layer-panel ${this.props.className}`}
          style={this.props.style}
          onMouseDown={this.props.onMouseDown}
          onTouchStart={this.props.onTouchStart}
        >
          <WidgetPanelHeader
            isConfigActive={isConfigActive}
            id={widget.id}
            idx={idx}
            isVisible={config.isVisible}
            label={config.label}
            labelRCGColorValues={datasets[config.dataId].color}
            widgetType={widget.name}
            onToggleEnableConfig={this._toggleEnableConfig}
            onToggleVisibility={this._toggleVisibility}
            onUpdateWidgetLabel={this._updateWidgetLabel}
            onRemoveWidget={this._removeWidget}
          />
          {isConfigActive && (
            <WidgetConfigurator
              layer={widget}
              lidgets={layers}
              datasets={datasets}
              layerTypeOptions={widgetTypeOptions}
              openModal={this.props.openModal}
              updateLayerConfig={this.updateWidgetConfig}
              updateLayerVisualChannelConfig={
                this.updateWidgetVisualChannelConfig
              }
              updateLayerType={this.updateWidgetType}
              updateLayerVisConfig={this.updateWidgetVisConfig}
            />
          )}
          {config.isVisible && config.aggregatedData ? (
            <StyledWidgetContent>
              {config.aggregationType}-{config.aggregatedData}
            </StyledWidgetContent>
          ) : null}
        </PanelWrapper>
      );
    }
  }

  return WidgetPanel;
}

export default WidgetPanelFactory;
